( -*- forth -*- )

( example app for 14bit forth. tested on 12f629 )

( alternative design to synth14.f using timer1 to switch
the output on or off. this gives better time resolution than
a fixed clock, but needs 16bit math.

a much better way of course is to use a pic with
capture/compare to do 'analog' pwm. this leaves timer1 as
main osc period, and the compare as the duty cycle.

high frequency pwm can be done using true pwm output
to generate other kinds of waveforms.

another option is to use several timers on multi-output
pic: one for the main osc period [on], the others for multi
pin outputs and square wave hard sync.

)


( should lead to digital/analog hibrid.
pic used for:
* control signal generation
* waveform generation?

oscillators follow this scheme:

[A]----/\/\/\----[B]----O
                  |
                 ===
                  |
                  |
                  V

pin [A] can be pulsed between 0/1 and Z, with
B set to Z, to charge the cap incrementally.

pin [B] can be set to 0/1 with [A] set to Z,
to discharge/charge the cap.

the Z [input] modes should be analog to prevent
pic digital input stage excessive current consumption.


tis can be used to generate a lot of control signals.




)

load stc14/stc.f
load stc14/forth.f


target
( pwm periods )
variable up
variable uph
variable down
variable downh

( global period var)
variable full
variable fullh

host

( this is the instruction rate )
1000000 constant cpufreq

( convert frequency to oscillator period: freq -- period )
: hz        cpufreq swap / ;



( INIT )

macro

: timer1
    0x5 T1CON ! ;  ( enable, no prescaler, internal clock )

: output5
    bank1 0xf TRISIO ! bank0 ;  ( GP5 = out, rest = in )


( OSCILLATOR )

( pin status )
macro
: toggle  0x20 GPIO xor! ;
: on?     GPIO 5 high? ;
: off?    GPIO 5 low? ;
: on      GPIO 5 high ;
: off     GPIO 5 low ;


target

: up@  up  @ uph  @ ;
: down@ down @ downh @ ;

variable state

state machine : period@ up@ down@ ; 

: period  1 state xor! period@ ;

: schedule ( lo hi -- )
    bank0 PIR1 TMR1IF low     ( clear interrupt bit )
    bank1 PIE1 TMR1IE high    ( enable timer interrupt )
    bank0
    [ INDF comf, TMR1H movwf, drop, ] ( set timer value )
    [ INDF comf, TMR1L movwf, drop, ]
    iret ;
    

( main isr: routes to one of the lower freq counters
all isr routines call 'done' at the end to set the timer
execution: uses 2 return stack cells )

interrupt : sample toggle period schedule ;


load stc14/math.f

target
: pwm
    up down full add3
    full b move
    a=frac(b)
    a up move ;
    

( MAIN LOOP )

application : main
    ( setup )
    clearmem digital output5
    timer1 interrupts
    bank1 0x80 OPTION_REG !
    bank0
    [ 1000 hz lohi ]  lit uph  ! lit up ! 
    [ 1000 hz lohi ] lit downh ! lit down !
    0 state !
    sample 
    begin
	dup pwm
    again ;


there 2 / .x cr
savehex synth14b.hex



