( -*- forth -*- )
( 18fxx2 instructions )


: ANDWF  0x1400 ;
: ADDWF  0x2400 ;
: ADDWFC 0x2000 ;

: SUBWF  0x5C00 ;
: SUBWFB 0x5800 ;
: DECF   0x0400 ;
: INCF   0x2800 ;
: XORWF  0x1800 ;
: IORWF  0x1000 ;

: MOVWF  0x6E00 ;
: MOVF   0x5000 ;
: MOVLW  0x0E00 ;
: MOVFF  0xC000 ;

: ADDLW  0x0F00 ;
: SUBLW  0x0800 ;
: ANDLW  0x0B00 ;
: XORLW  0x0A00 ;
: IORLW  0x0900 ;


: PUSH   0x0005 ;
: POP    0x0006 ; 

: BZ     0xE000 ;
: BOV    0xE400 ;
: BNOV   0xE500 ;
: BNN    0xE700 ;
: BNC    0xE300 ;
: BN     0xE600 ;
: BC     0xE200 ;
: BNZ    0xE100 ;
: BRA    0xD000 ;
: RCALL  0xD800 ;

: NOP    0x0000 ;
: NOPF   0xF000 ;   ( nop used as extension word for 2 word instruction )
: RETURN 0x0012 ;
: RETFIE 0x0010 ;
: RETLW  0x0C00 ;
: CALL   0xEC00 ;
: GOTO   0xEF00 ;

: LFSR   0xEE00 ;

: TBLRD*  0x0008 ;
: TBLRD*+ 0x0009 ;
: TBLRD*- 0x000A ;
: TBLRD+* 0x000B ;
: TBLWT*  0x000C ;
: TBLWT*+ 0x000D ;
: TBLWT*- 0x000E ;
: TBLWT+* 0x000F ;


: RESET   0x00ff ;

: RLCF    0x3400 ;
: RLNCF   0x4400 ;
: RRCF    0x3000 ;
: RRNCF   0x4000 ;

: BCF     0x9000 ;
: BSF     0x8000 ;
: BTG     0x7000 ;
: BTFSC   0xB000 ;
: BTFSS   0xA000 ;

: SWAPF   0x3800 ;

: CLRF    0x6a00 ;
: SETF    0x6800 ;
: COMF    0x1C00 ;
: NEGF    0x6C00 ;

: RETFIE1 0x0011 ;
: SLEEP   0x0003 ;

: CPFSGT  0x6400 ;
: CPFSLT  0x6000 ;
: CPFSEQ  0x6200 ;

( forth opcodes )
( stacks grow upward due to absence of PREDEC autoincrement modes )
: DROP POSTDEC0 0xff and MOVF or ;           ( w = *FSR0-- )
: DUP  PREINC0 0xff and MOVWF or 0x200 or ;  ( *++FSR0 = w )


( remarks:
MOVFF cannot use PCL, TOSx as destination reg
MOVFF, MOVLW and MOVWF do not affect flags
MOVF  affects flags N,Z flags, not carry. can be used for testing )

( addressing modes )
( to enable defaults and simplify the asm, the d and a bits are modified
in the target buffer so we have

TOSL addwf, f, b,

instead of

TOSL f b addwf, )

: f, ( -- ) prev t@ 0x0200 or prev t! ;  ( dst = freg )
: b, ( -- ) prev t@ 0x0100 or prev t! ;  ( use bsr instead of access bank )
: fop, ( reg opcode -- ) swap 0xff and or , ;
: ffop, ( src dst opcode -- ) >r swap 0xfff and r> or , 0xfff and NOPF or , ;




( assembler ops / target macros taking arguments )

: movwf,  ( dst -- )      MOVWF fop, ;
: movf,   ( src -- )      MOVF fop, ;
: movlw,  ( lit -- )      MOVLW fop, ;
: movff,  ( src dst -- )  MOVFF ffop, ;
: retlw,  ( lit -- )      RETLW fop, ;

: lfsr,   ( value reg --  )  ( load FSR register with 2 byte literal )
    >r dup
    8 >> 0x0f and LFSR r> 4 << or or , ( compile first word )
    0xff and NOPF or , ;               ( compile second word )

: nop,  NOP , ;
: push, PUSH , ;
: pop,  POP , ;
: reset, RESET , ;


( nop messages )
: bad,     there 0xfbad , 0xbad .x .x ;


( bit operations )

( reg bit opcode -- opcode )
: bitop,
    >r 0x7 and 9 << r> or   ( bit operand + opcode )
    swap 0xff and or , ;    ( reg operand )

( reg bit -- )
: bsf,   BSF bitop, ;
: bcf,   BCF bitop, ;
: btg,   BTG bitop, ;
: btfsc, BTFSC bitop, ;
: btfss, BTFSS bitop, ;


( src -- )
: addwf, ADDWF fop, ;
: andwf, ANDWF fop, ;
: subwf, SUBWF fop, ;
: subwfb, SUBWFB fop, ;
: iorwf, IORWF fop, ;
: xorwf, XORWF fop, ;

( imm -- )
: addlw, ADDLW fop, ;
: andlw, ANDLW fop, ;
: sublw, SUBLW fop, ;
: iorlw, IORLW fop, ;
: xorlw, XORLW fop, ;

( reg -- )
: incf,     INCF fop, ;
: decf,     DECF fop, ;
: rrcf,     RRCF fop, ;
: rlcf,     RLCF fop, ;
: rlncf,    RLNCF fop, ;
: rrncf,    RRNCF fop, ;
: setf,     SETF fop, ;
: clrf,     CLRF fop, ;
: negf,     NEGF fop, ;
: comf,     COMF fop, ;
: cpfsgt,   CPFSGT fop, ;
: cpfslt,   CPFSLT fop, ;
: cpfseq,   CPFSEQ fop, ;



: swapf, SWAPF fop, ;


: retfie,   1 and RETFIE or , ;

( branching )

: call?      2prev t@ 0xf000 and CALL = 2last? and ;
: rcall?     prev t@ 0xf800 and RCALL = last? and ;
: atailcall, 2prev t@ 0xfff and GOTO or 2prev t! peephole pop drop ;
: rtailcall, prev t@ 0x7ff and BRA or prev t! peephole pop drop ;


: blong,    ( operands opcode -- ) ( compile a 2word long branch instruction )
    >r 2 / lohi swap               ( save opcode and convert to word addressing )
    r> or , NOPF or , ;


( - 2 to compensate for PC inc, 2 / to use word addressing )
: encode-offset  ( distance -- rel.encoding ) 2 - 2 / ;
: decode-offset ( rel.encoding -- distance ) 2 * 2 + ;

( to from -- relative.encoding )
: relative  - encode-offset ;

: offset-to    there relative  ;              ( compute branch offset there->address )
: offset-from  there swap relative ;          (                       address->there )
: rjump,       offset-to 0x7ff and BRA or , ;      ( short relative jump )
: rcall,       offset-to 0x7ff and RCALL or , ;    ( short relative call )


( relative.encoding -- ? ) 
: fits?  0x400 + 11 >> not ; ( for BRA and RCALL )
: cfits? 0x80  + 8  >> not ; ( for conditional branches )

( address -- ? )
: long?  dup there relative fits? not ;   ( can't use BRA )
: clong? dup there relative cfits? not ;  ( can't use conditional branch )

( absolute [far] jumps )
: ajump,  GOTO blong, ;
: acall,  CALL blong, ;
    
: jump,   long? if ajump, else rjump, then ;
: call,   long? record if acall, else rcall, then ;



( conditional branching )
( address opcode -- )
: short-cjump, >r offset-to 0xff and r> or , ;
: patch-cjump, ( addr )
    dup offset-from 0xff and >r  ( compute relative distance ) 
    dup t@                       ( get opcode )
    0xff00 and r> or             ( strip address and insert new )
    swap t! ;                    ( store instruction )

( address -- packed opcode )
: short-jump, offset-to 0x7ff and BRA or ;


( address opcode -- )
: long-cjump,
    there 4 + swap   ( address past the long jump )
    swap 0x0100 xor  ( reverse polarity of conditional jump )
    short-cjump,     ( pack opcode and compile )
    GOTO blong, ;    ( compile the long jump )

( compile short conditional jump, or its
inverse which jumps over a long jump )

( address opcode -- )
: cjump, >r clong? if r> long-cjump, else r> short-cjump, then ;



( load disassembler )
load asm/dasm16bit.f