( -*- forth -*- )

( 16bit DTC from scratch : alternative to 8bit STC 

IP = table pointer
DS = *FSR0
RS = native rs
A  = FSR2 : indirect addressing

FSR1 unused )

: kernel  0x40 ;      ( start compiling kernel here )
: data    0x7f ;      ( initial data stack ptr )

( for math ops 1->1 and 21->1 )
: x1     0x0 ; ( 2ND )
: x0     0x1 ;
: y1     0x2 ; ( TOP )
: y0     0x3 ;

( STAGE 1 )

load host/compiler.f ( host additions )
load arch/18f452.f   ( control register names )
load asm/16bit.f     ( basic 16bit core assembler )
load dtc/native.f    ( native code compiler for target primitives )

( macros for below. those containing = are register macros:
they do not push/pop the stack implicitly, and have C like names )
load dtc/mreg.f   ( register style macros, names resembling C syntax )
load dtc/mforth.f ( forth style macros )

kernel there! target
load dtc/pfactors.f       ( primitive native code factors )
load dtc/primitives.f     ( primitive words )

( STAGE 2 )

host
load dtc/threaded.f  ( second stage bootstrap: from here on where're DTC big endian )
load arch/m18f452.f  ( reload macros after lit is redefined )
load dtc/forth.f     ( highl level forth )
load dtc/mem.f       ( interpreted target memory management words )
load dtc/boot.f      ( dtc boot and interrupt code )


load host/view.f     ( make target look like a standard forth )



( -*- forth -*- )

( REWORK AS: )

( attempt to write a 16bit [near]ansi forth in terms
of the 8bit forth for the 16bit core. this is supposed
to be an standard abstraction layer hiding most of pic's
architecture.

words are defined in lowercase, but interpreter will be
patched so uppercase words are possible.

forth map:

all data is stored in little endian format, stacks
grow upward.

there is a single 16bit unified memory model, which
is word addressed.

0x0000 - 0x07ff  DATA + SFR  [ 0x00000 - 0x00fff ]
0x0800 - 0x0fff  EEPROM      [ 0x01000 - 0x01fff ]
0x1000 - 0xffff  PROGRAM     [ 0x02000 - 0x1ffff ]

the native code kernel resides in the non-accessible
part below 0x1000 [ 0x2000 ]

eventual goal is to have a native forth, with dictionary
and everything running on the target.

)


