( -*- forth -*- )

( BADNOP dev env for PICmicro 18fxx2 )

( note on name attributes:

compiler macros end with '_'
target macros end with ','
target commands end with '#'

)

( ********** COMPILER EXTENSION  ************ )


( target dictionary and code buffer )
: <t'>     accept tfind ;
: [t']_    <t'> lit_ ;

variable code-buf
variable code-free
variable code-mark ( target address )
variable code-address-mask
variable code-word-mask
0xffff code-word-mask h!


: t>h ( targ.addr -- host.addr )
    code-address-mask h@ and
    code-buf h@ + ;

: there  code-free h@ code-buf h@ - ;
: there! t>h code-free h! ;

    

: tcode  ( bytes -- )
    dup 1 - code-address-mask h!
    here code-buf h!
    here code-free h!
    there code-mark h! ( target units )
    3 + 4 / -table ;    ( FIXME )



: tmask   code-word-mask h! ;
: (tmask) code-word-mask h@ and ;
: tc!    t>h hc! ;     
: tc@    t>h hc@ ;
: t!     >r (tmask) dup r tc! 8 >> r> 1 + tc! ;  ( target machine is little endian )
: t@     >r r tc@ r> 1 + tc@ 8 << or (tmask) ;

( attributes )
: a_tmacro   0x2c ;
: a_tcommand 0x23 ;
: a_lword    0x26 ;

: tword    accept there (tdict) ;
: (tmacro)   a_tmacro   attrib! (hword) unattrib ;
: (tcommand) a_tcommand attrib! (hword) unattrib ;
: (lword)    a_lword    attrib! (hword) unattrib ;

: tmacro   accept (tmacro) ;
: lword    accept (lword) ;

: , there dup >r t! r> 2 + there! ;   ( x -- ) ( compile )


: tmark  there code-mark h! ;
: mark   hmark tmark ;

: tempty code-mark h@ there! ;
: empty  hempty tempty ;

( source file re-loading
this is a bit tricky, since we overwrite the image, so we go like:
* load pointer to name on DS. name string is in host buffer.
* empty. after this, the name string resides in 'free' memory
* load the file. before any new code is defined, the name is used,
  so it can't be overwritten yet. )

: [empty-load]_ [name]_ empty_ (load)_ ;




( peephole optimization for certain instructions:
-- stack movement optimization: makes sense since usually dup/drop is an explicit instruction
-- tail call optimisation: necessary for proper usage of tail recursion
-- literal folding.

the optimizer is very simple: just emit code for each word in a colon definition
if there is no possible optimization, else undo a previous compilation and combine
it's effect with the current one. all optimizable instructions are recorded
on the peephole stack. )

stack peephole

: prev     there 2 - ; ( previous instruction )
: 2prev    there 4 - ; ( previous double instruction )
: 3prev    there 6 - ;



: record   there peephole push ;    ( mark an instruction for opti )
: last?    peephole top prev  = ;   ( previous instruction is marked? )
: 2last?   peephole top 2prev = ;   ( .. double .. )
: 3last?   peephole top 3prev = ;
: rewind   peephole pop there! ;    ( undo last instruction )

: undo     prev there! ;





( swap cells in target memory )
: tswap ( a1 a2 -- )
    over t@ over t@      ( a1 a2 d1 d2 )
    >r swap t! r> swap t! ;

: t2swap ( a1 a2 -- )
    over 2 + over 2 + tswap tswap ;



: checkme there 2 / .x space 3 for [c] ! emit next cr ;
: marker tword ;


( file io )

load host/hex.f ( hex file output )





