( -*- forth -*- )

( part 2 of badnop: extensions to kernel [badnop.c]
the idea is that:
* forth.f    is always loaded. provides basic host forth. terminated by mark.
* compiler.f is [re]loaded by the target forth code )

( note on addressing:
compiler h!, h@ use real processs addresses.
target t!, t@ use virtual addresses )

( *********** HOST FORTH EXTENSIONS ************* )

( parsing )

: hfind   (hfind) drop ;
: tfind   (tfind) drop ;

: <h'>     accept hfind ;
: [h']_    <h'> lit_ ;

: hword    accept (hword) ;
: number   accept (number) ;
: c        accept buffer h@ 0xff and ;
: [c]_     c lit_ ;
: space   32 emit ;
: tab      9 emit ;


( words )
: not -1 xor ;
: neg not 1 + ;
: nip swap drop ;
: ]L  lit_ ] ;
: create   hword dovar_ ;     ( create dict entry and compile dovar's code field )
: constant hword doconst_ _ ; ( .. doconst .. )


( host macros )
: for_      >r_ here ;
: next_     donext_ _ rdrop_ ;
: variable  create 0 _ ;                   ( initialized variable )
( : symbol   create here accept readword ;  ( read next word and store as symbol )
: table     for 0 _ next ;                  ( initialized table )
: -table    for -1 _ next ;                 ( all bits set )
: 0jump_    here 0 _ ;                      ( load address and reserve space )
: if_       jz_ 0jump_ ;
: then_     here swap h! ;                  ( backpatch jump address )
: else_     jmp_ 0jump_ swap then_ ;
: begin_    here ;
: again_    jmp_ _ ;
: while_    if_ ;
: repeat_   swap again_ then_ ;
: until_    not_ while_ repeat_ ;

( parsing macro: inserts call on return stack )
: atexit_   r>_ <h'> 1 cells + lit_ >r_ >r_ ;
    
: h+!       >r r h@ swap + r> h! ;
: cell      1 cells ;
: 0=        0 = ;
: 1+        1 + ;
: 1-        1 - ;

: forstring_  begin_ dup_ hc@_ ;
: forchar_    accept_ buffer_ forstring_ ;
: nextchar_   1+_ dup_ hc@_ 0=_ until_ drop_ ;
: echo        forchar emit nextchar ;


: upper? ( c -- ? ) dup [c] A >= swap [c] Z <= and ;
: lower? ( c -- ? ) dup [c] a >= swap [c] z <= and ;

: upper ( c -- C )  dup lower? if [ c A c a - ]L + then ;
: lower ( c -- C )  dup upper? if [ c a c A - ]L + then ;

: cell+ 1 cells + ;

: .string     forstring emit nextchar ;
: .lowstring  forstring lower emit nextchar ;


: [name]_
    here 4 cells + lit_ jmp_ 0jump_
    forchar char_ nextchar 0 char_ align then_ ;

: [echo]_ [name]_ .string_ ;
: [load]_ [name]_ (load)_ ;
: [output-file]_ [name]_ (output-file)_ ;


( stack: bottom contains top pointer 
stacks grow upward in memory
popping an empty stack returns 0
pushing a full stack resets the stack, then pushes )

: default-size 256 ;            ( default stack size )
: stack		        	( create a stack )
    create 			( create tos pointer )
    here _                      ( make TOS point to itself: empty stack )
    default-size table ; 	( create stack buffer )

: reset    dup h! ;		   ( make TOS point to itself )
: elements dup h@ swap - cell / ;  ( get nb of elements on stack )
: empty?   elements 0 = ;
: full?    elements  default-size = ;

: top      dup empty? if drop 0 exit then h@ h@ ;


: push
    dup full? if dup reset then ( FIXME: should abort )
    dup >r			( save address of stack pointer )
    h@ cell +                   ( deref and pre-inc )
    dup r> h!	                ( store new value of stack pointer )
    h! ;                        ( store value on new TOS )

: pop
    dup empty? if drop 0 exit then
    dup >r                      ( save address of stack pointer )
    h@ dup >r                   ( get TOS ptr + save value )
    h@                          ( get value on TOS )
    r> cell - r> h! ;           ( post-dec stack pointer and store it )
	


( these are redefined in compiler.f to include
target code buffer )
: mark hmark ;
: empty hempty ;


( misc util )
: lohi dup 0xff and swap 8 >> 0xff and ; ( word16 -- low8 high8 )
: 2dup over over ;
: 2drop drop drop ;

( vectored code: parses variable name
stores code xt after name in variable )
: vector_
    here 5 cells + lit_  ( 2 )
    <h'> _               ( 1 )
    h!_ exit_            ( 2 )
    enter_ ;

( hex digits )
: >hex ( digit -- ascii ) 0xf and 10 over > if [c] 0 + else [c] A 10 - + then ;
: hex> ( ascii -- digit ) dup [c] A >= if 10 [c] A - + else [c] 0 - then 0xf and ;


: MHz 1000000 * ;


: max 2dup > if drop else nip then ;
: min 2dup < if drop else nip then ;




( keyboard input
this should eventually replace readline )

: key -1 charmode 0 read-fd 0 charmode ; ( -- ascii )
