( -*- forth -*- )

( boot code for stc14
this is different from the 1tc16, because we have dup as
a function, not a macro, so it's easier to have the boot code
and dup defined up front. )


( WREG macros: don't push/pop the stack )
: top=w,  INDF movwf, ; 
: w=top,  INDF movf, ; 
: dp=w,  FSR movf, ;
: w=dp,  FSR movwf, ;

( FSR )
: dp++,  FSR incf, f, ;
: dp--,  FSR decf, f, ;
: drop,  dp++, ;


( start compiling boot code at 0 )

( reset code )
target 0 there!
0 jump,      ( 0x000 ) 
0 jump,      ( 0x001 )
0 jump,      ( 0x002 ) 
0 jump,      ( 0x003 ) 
0 jump,      ( 0x004 ) 

host

( execution tokens )
variable xtfree
0x8 xtfree h!
: xt-current ( -- xt )   xtfree h@ ;
: xt-check               xt-current 2 * kernel > if e-page then ;
: xt-allot   ( n -- xt ) xtfree h+! xt-check ;
: xt-there!  ( -- prev ) there xtfree h@ 2 * there! ;

: xt[_ xt-there!_ >r_ ;
: ]xt_ r>_ there!_ 1 lit_ xt-allot_ ;


: xt! ( addr xt -- ) there >r 2 * there! jump, r> there! ;
: xt# ( addr -- ) xt-there! swap jump, there! 1 xt-allot ;

( start compiling past xt table )
target kernel there!

there : dup     w=top ; undo
there : push=w  dp-- top=w ; 

host
( dup and push=w are fake opcodes, but we put them
in a constant for optimization purposes )
: fake-opcode 2 / 0x3ff and CALL or ;

fake-opcode constant PUSH=W
fake-opcode constant DUP
0x0A84      constant DROP
0x0080      constant TOP=W

: dup?    prev t@ DUP = ;
: drop?   prev t@ DROP = ;
: push=w? prev t@ PUSH=W = ;
: top=w?  prev t@ TOP=W = ;


: application there 0 xt! ;
: interrupt   there 4 xt! ;



( redefine compile, to include push/pop optimisation )
: dst-pop? >r ( address -- address ? )
    r t@     0x0800 ( INDF movwf, ) =
    r 2 + t@ 0x0a84 ( FSR decf, f,) = and
    r> swap ;

( if called function starts with w=pop, we can skip a previous
push=w and jump into the function after the w=pop sequence )
: compile,
    dst-pop? push=w? and landing?
    not and if undo 4 + then call, ;

: default-compile  vector <compile,> compile, ; 
: :# default-exit default-compile tword ] ;



( redefine drop to eliminate dup and push=w
should only occur due to the effect of macros )

: drop, dup? push=w? or if
	undo ( there .x [c] D emit cr ( i'm curious if this ever happens.. )
    else dp++, then ;

: w=pop,
    push=w? if undo exit then
    INDF movf, drop, ; 
