( -*- forth -*- )
( fetch/store )

: writer? ( address -- ? )
    t@ 0x3f80 and MOVWF = ;


: @,  <lit, movf, push=w, ;  ( fetch/store only use literal addresses )
: !,
    <lit,                 ( address needs to be literal )
    lit? if <lit,         ( argument optionally literal )
	dup 0 = if        ( if zero, compile clrf )
	    drop clrf,
	    exit then
	movlw, movwf,
	exit then
    dup? 2prev writer? and if ( dup followed by writer )
	undo movwf, exit then
	
    w=top, drop, movwf, ; ( non literal argument )

: @?
    2prev t@ 0xff80 and MOVF = push=w? and ;

: <@,
    @? if 2prev there! there t@ 0x7f and
	exit then e-nolit ;


( partly implemented: only regster args )
: cmp,
    <@, lit? if <lit, movlw, subwf, exit then ( reg lit )
    <@, movf, subwf, ;  ( 2 reg args )

: <?,  cmp,  c?, ;
: >=?, cmp, nc?, ;
: =?,  cmp,  z?, ;



: 0!, <lit, clrf, ;

( so we don't have to optimize @ followed by ! this )
: @!, <lit, <lit, movf, movwf, ;

: high, <lit, <lit, swap bsf, ;
: low,  <lit, <lit, swap bcf, ;


: lit-op, ( opcode value -- ) movlw, INDF swap fop , f, ;

: op, <lit, lit-op, ;

( optimized version:
if we have a push, we can delay that, and operate
on wreg directly.

difference = 0x3c00
ANDLW ANDWF
IORLW IORWF
XORLW XORWF

)

: logic-op,
    <lit, push=w? not if lit-op, exit then              ( no push -> default )
    undo swap [ ANDLW ANDWF xor ]L xor lop , push=w, ;  ( push -> squeeze inbetween )
    

( : logic-op, op, ; ( )

: and,  ANDWF logic-op, ;
: or,   IORWF logic-op, ;
: xor,  XORWF logic-op, ;
: not,  INDF comf, f, ;

( target : x 123 @ 7 and 7 and 7 and ; host ( )



: -,    SUBWF op, ;
: +,    ADDWF op, ;

: op!,
    <lit,
    lit? if <lit, movlw, swap fop , f, exit then
    @?   if <@,   movf,  swap fop , f, exit then
    
    INDF movf, swap fop , f, drop, ;

: +!,   ADDWF op!, ;
: -!,   SUBWF op!, ;
: or!,  IORWF op!, ;
: xor!, XORWF op!, ;
: and!, ANDWF op!, ;

: <<!,  RLF op!, ;


( misc macros )

: clc,   STATUS C bcf, ;
: stc,   STATUS C bsf, ;
: c>>r,  INDF rrf, f, ;
: r<<c,  INDF rlf, f, ;
: >>,    INDF clc, rrf, f, ;
: <<,    INDF clc, rlf, f, ;
: 2*,    <<, ;
: >>r,   clc, INDF rrf, f, STATUS C btfsc, INDF 7 bsf, ;
: 2/,    clc, INDF 7 btfsc, stc, c>>r, ;

: 1-,   INDF decf, f, ;
: 1+,   INDF incf, f, ;


: 1+!,   <lit, incf, f, ;
: 1-!,   <lit, decf, f, ;

: swapdigits, INDF swapf, f, ;

: c>>!,   <lit, rrf, f, ;


( optimizations
the most important one still not implemented is
top of stack caching, which is not so simple to
implement compared to the other optimzations, but
which can have quite an impact on math ops.

one example, 'reg @ 0x7 and route'

MOVF 0x32, W
CALL 0xC       [push=w]
MOVLW 0x7
ANDWF INDF, F
MOVLW 0
MOVWF PCLATH
MOVF INDF, W
INCF FSR, F
ADDWF PCL, F


this is quite something, and can be reduced to

MOVLW 0
MOVWF PCLATH
MOVF 0x32, W
ANDLW 0x7
ADDWF PCL, F

but doing this automaticly will involve quite
a lot of shuffling, since the loading of PCLATH
needs to move up, so all math can be done in wreg.

)