( -*- forth -*- )

( boot & interrupts )


( interrups call either the low or high priority handler.
we provide global handlers for both with proper context
save/restore, so specific interrupt handlers can be ordinary
forth code. the virtual interrupt table is a series of
bittest + skip instructions.


)

host
variable chainhi
variable chainlo
variable chaininit


target


( can be done simpler
high pri interrupt does not need saving:
when a is modified, it needs to be saved in the handler.
these are supposed to be fast. dispatching can be done
inline. )



( this is the end of the dispatcher chain )
: exithi [ (exithi),
: exitlo [ (exitlo),

: enterhi [
    push,                ( set return address )
    <t'> exithi lohi
    movlw, TOSH movwf,
    movlw, TOSL movwf,
    (enterhi),
    there chainhi h!     ( record vector )
    0 ajump,             ( default handler is reset )
    
: enterlo
    (enterlo)                 ( save context )
    push [ <t'> exitlo lohi ] ( set return address )
    lit TOSH ! lit TOSL !     
    [ there chainlo h!
    0 ajump,                  ( default handler is reset )

: init [
    initstacks
    there chaininit h!        ( record vector )
    reset, nop,               ( terminator = reset )
    
host
    
( compile self-linked dispatcher link
  will be patched into link using a GOTO swap )
   
    
( address reg bit -- jump )
: flaglink,
    there >r               ( save where we're at )
    btfsc, jump,           ( compile test + jump )
    there r> ajump, ;      ( compile jump to self )

( add a bit tester to a dispatcher chain.
  the chain pointers are 2 cell GOTO instructions )

( address reg bit tailvar -- )
: >flagchain
    >r flaglink, r>     ( jump tailvar )
    2dup h@ t2swap      ( swap GOTO with head's = add to chain )
    h! ;                ( save new tail )
    

( add a handler to the chain )
( address reg bit -- )

: >ihigh chainhi >flagchain ;
: >ilow  chainlo >flagchain ;
    
( address -- jump )
: link,   there >r call, there r> ajump, ;
( address tailvar -- )
: >chain  >r link, r> 2dup h@ t2swap h! ;

( add an address to the init chain
  maybe it's best to call inits directly from main program )
: >init   chaininit >chain ;

( set end of init chain to point here )

: terminal ( chain -- )
    there >r h@ there! r jump, r> there! ;
    

: application chaininit terminal ;
    
    

( block0 = execution tokens: 256 vectored words
  note that long jumps take 2 slots
  the first 16 xts are for the monitor  
    
    0x0000 [ 0x00 ] reset    
    0x0002 [ 0x01 ] erase    
    0x0004 [ 0x02 ] flash    
    0x0006 [ 0x03 ] execute  
    0x0008 [ 0x04 ] inthi    
    0x000A [ 0x05 ] jsr      
    0x000C [ 0x06 ] emit     
    0x000E [ 0x07 ] a!       
    0x0010 [ 0x08 ] @a+      
    0x0012 [ 0x09 ] !a+      
    0x0014 [ 0x0A ] p!       
    0x0016 [ 0x0B ] @p+      
    0x0018 [ 0x0C ] intlo    
    0x001A [ 0x0D ] bye      
    0x001C [ 0x0E ] lit
    0x001E [ 0x0F ] ack
)    

    there 0x0 there!
    <t'> init jump,      ( 0x00 )
    0 jump,              ( 0x01 )
    0 jump,              ( 0x02 )
    0 jump,              ( 0x03 )
    <t'> enterhi jump,   ( 0x04 ) 
    0 jump,              ( 0x05 )
    0 jump,              ( 0x06 )
    0 jump,              ( 0x07 )
    0 jump,              ( 0x08 )
    0 jump,              ( 0x09 )
    0 jump,              ( 0x0a )
    0 jump,              ( 0x0b )
    <t'> enterlo jump,   ( 0x0c ) 
    0 jump,              ( 0x0d )
    0 jump,              ( 0x0e )
    0 jump,              ( 0x0f )
    there!
    

	
	



