( -*- forth -*- )


( conditional blocks use the machine flags
since handling flags & jumps are a big part of what a controller needs to do,
it might be nice to introduce the route and forward idioms, and leave if-then-else
for cases where this is not necessary. if[flag] should work, but is hard to see
because it's not always clear exactly which istruction is being used.

route [ ... x cond -- ] and forward [ ... x cond -- x ] pass to other words
if a coundition is met. the condition generator might be optimizable into
one of the many conditional branch and skip instructions.

condition generators have the stack effect [ -- flag ] so they don't drop
anything.

)


( using the return stack
: for,
    >r, there        
    peephole reset ;    
: next,              
    rl--,
    BNZ cjump,
    POP , ;
( end )

( using the temp stack
this version drops before the loop: better for
code which uses the top of stack )

: xfor,   there peephole reset ;
: for0,   >x, xfor, ;
: next0,  x-, BNZ cjump, xdrop, ;

( drop..save around loop code seems to generates better code:
most loops do "read modify write", which makes them start with
save and end with drop, so both are cancelled out. )
: for1,   dup, >x, xfor, drop, ;
: next1,  save, x-, BNZ cjump, xdrop, drop, ;


( intelligent next. if the address we jump to is a drop,
we patch the address to jump past it [turns the for1 into a for0],
and compile next0, else we compile next1 )
: nexti,  dup t@ DROP = if 2 + next0, exit then next1, ;

: next,   ( addr xt -- ) execute ;
: for,
    lit00? litff? or if for0, [h'] next0, exit then  ( override dup/save opti for 0x00 / 0xff lit counts )
    for1, [h'] nexti, ;                              ( default: use intelligent next with dup/save opti )

( these should be idiomized )
: ifnz,  there BZ , ;
: ifz,   there BNZ , ;
: ifc,   there BNC , ;
: ifnc,  there BC , ;
: ifn,   there BNN , ;
: ifnn,  there BN , ;
: ifov,  there BNOV , ;
: ifnov, there BOV , ;
: skip,  there BRA , ;


( compliment a jump instruction in target memory )
: jump-toggle ( opcode -- ) 0x0100 xor ;




  
( machine flag normalized conditon generators
  these can be optimized by if into conditional jumps )

: z?,   STATUS Z 1 cond, ;
: nz?,  STATUS Z 0 cond, ;
: c?,   STATUS C 1 cond, ;
: nc?,  STATUS C 0 cond, ;
: n?,   STATUS N 1 cond, ;
: nn?,  STATUS N 0 cond, ;
: ov?,  STATUS OV 1 cond, ;
: nov?, STATUS OV 0 cond, ;

macro
: =   xor nfdrop z? ;

( unsigned comparison )
: >=  -   nfdrop c? ;   
: <   -   nfdrop nc? ;

host


( using bult in compare/skip ins )
: =?, ;

( TODO: redefine if, for optimization

cond? no  -> ">flags ifz,"
STATUS    -> conditional jump
other     -> skip + branch

)

variable reg
variable bit
variable 0/1

: (if),    >flags, ifnz, ;  


: if,
    cond? not if (if), exit then   ( not optimizable )
    <cond, 0/1 h! bit h! reg h!    ( undo cond, )

    ( STATUS conditions -> conditional jump instructions )
    reg h@ STATUS 0xff and = if
	bit h@ Z  = if 0/1 h@ if ifz,  else ifnz,  then exit then
	bit h@ N  = if 0/1 h@ if ifn,  else ifnn,  then exit then
	bit h@ OV = if 0/1 h@ if ifov, else ifnov, then exit then
	bit h@ C  = if 0/1 h@ if ifc,  else ifnc,  then exit then
    then
    ( other conditions -> conditional skip + branch )
    reg h@ bit h@                ( reg bit )
    0/1 h@ 1 xor 12 <<           ( compute set/clear bit )
    BTFSS or                     ( reg bit BTFSS/BTFSC )
    bitop, skip,
;


( then optimizations )

( short "exit then" -> look for jump and invert if possible

conditions:
* bra instruction
* argument needs to fit
assumptions:
* bra is backwards [ only older words ]

)

( opcode -- ? )
: op-cbra?   0xF800 and 0xE000 = ;
: op-bra?    0xF800 and BRA = ;
: op-rcall?  0xF800 and RCALL = ;
: op-btfs?   0xE000 and BTFSS = ;

( -- ? )
: cbra?  prev t@ op-cbra? ;
: -bra?  prev t@ 0xFF00 and 0xD700 = ; ( backwards jump with high 3 bit 111 )

( fetch address of last bra instruction )
( -- addr )
: <bra,
    -1 8 << prev t@ 0xff and or   ( use only low byte from opcode )
    decode-offset prev +          ( get absolute address )
    prev there! ;                 ( undo branch )


( conditional skip  -> "cskip bra thing ..." -> "ncskip thing ..." )


: (chainthen),
    >r <bra, r t@            ( undo branch, fetch cond branch instruction ) ( address opcode )
    0x0100 xor swap          ( reverse polarity ) ( opcode address )
    r relative 0xff and or   ( merge opcode )
    r> t! ;

( check if conditional branch can be combined
with a previous backwards relative branch.
this is to optimize 'if word exit then' )

( address -- address ? )
: jumpchain? >r
    r  t@ op-cbra?   ( do we have a conditional jump instruction? )
    r  2prev =       ( only one cell inbetween cjump and there? )
    -bra?            ( previous cell is 8bit negative branch? )
    and and r> swap ;

: (skipthen),
    drop
    3prev t@ 0x1000 xor 3prev t!  ( reverse polarity of btfs )
    prev t@ 2prev t!              ( move instruction )
    prev there!                   ( rewind )

    ( fix relative branches )
    prev t@ >r
    r 0xF000 and BRA =        ( moved instruction is BRA or RCALL ? )
    not if rdrop exit then    ( no -> done )
    r> 1 + prev t!            ( yes -> patch relative address )               
    
;

( address -- address ? )
: skipbra? >r
    r  t@ op-bra?     ( is the referred instruction a branch instruction? )
    r  2prev =        ( only one cell inbwetween bra and there? )
    3prev t@ op-btfs? ( is the instruction before that a btsf ins? )
    and and r> swap ;


: (then),                   ( can be used to patch any conditional reljump )
    patch-cjump, ;          ( it can patch forward bra too )

: then,
    peephole reset
    jumpchain? if (chainthen), exit then
    skipbra?   if (skipthen),  exit then
    (then), ;


: else,                     ( patch first cond jump and set new uncond )
    peephole reset
    skip, >r patch-cjump, r> ;




( ? -- )
: pass,      if, <t'> compile, exit, then, ;


( standard forth looping constructs )
: begin,     there peephole reset ;
: again,     jump, ;
: repeat,    swap again, then, ;
: while,     if, ;
: until,     not, while, repeat, ;

: do, there peephole reset ;

: (loop), INDF0 cpfseq, jump, drop, drop, ;

: loop, 1+, (loop), ;
: +loop, +, (loop), ;


macro
: abs       WREG 7 high? if neg then ;

host
