( -*- forth -*- )

host


( fast baud rate divisor )
: >brg 16 * fosc swap / 1 - ;


( see p18_tiri.asm AN774a )

target


( io buffers )
buffer rx
buffer tx

( give status bits a sane name )
PIR1 RCIF  2constant rx-int    ( got receive interrupt )
PIR1 TXIF  2constant tx-int    ( got transmit interrupt )
PIE1 RCIE  2constant rx-int-en ( rx interrupt enable )
PIE1 TXIE  2constant tx-int-en ( tx interrupt enable )
RCSTA OERR 2constant overrun
RCSTA FERR 2constant frame
RCSTA CREN 2constant rx-en

( save/restore implemented as macros )

: fatal    reset ;
: rx-reset rx-en low rx-en high ;

: rx-handler
    rx-int-en low?  pass fatal
    overrun high?   pass rx-reset
    RCREG @
    frame high?     if drop exit then 
    rx write ;



: tx-handler
    tx-int-en low?  pass fatal
    tx empty?       if tx-int-en low exit then
    read TXREG ! ;


: serialinit
    tx clear rx clear          ( clear buffers )
    TRISC 7 high TRISC 6 low   ( set tris bits for Tx and Rx )
    
    [ baud >brg ]L SPBRG !     ( set baud rate )
    0x24 TXSTA !               ( enable transmission and high baud rate )
    0x90 RCSTA !               ( enable serial port and reception )
    0x30 PIE1 or!              ( enable Tx and Rx interrupts )
    [ 0x30 not ]L IPR1 and!    ( set Rx Tx low priority )
    RCON IPEN high             ( enable interrupt priorities )
    0xc0 INTCON or!            ( enable global low and high ints )
;


( these do busy waiting
maybe it's best to save the a reg, to have the same
behaviour as serial_simple.f )

: transmit   tx begin room? until write tx-int-en high ;
: receive    rx begin ready? until read ;
    

( setup init and int handlers )
' rx-handler rx-int >ilow
' tx-handler tx-int >ilow
' serialinit >init

