( -*- forth -*- )

target
1 constant DEVICE
2 constant CONFIGURATION
4 constant INTERFACE
5 constant ENDPOINT

1 constant NUM_CONFIGURATIONS
1 constant NUM_INTERFACES

0x21 constant HID
0x22 constant REPORT


( Device )
0x12 , DEVICE ,    ( bLength, bDescriptorType )
0x10 , 0x01 ,      ( bcdUSB [low high] )
0x00 , 0x00 ,      ( bDeviceClass, bDeviceSubClass )
0x00 , 0x08 ,      ( bDeviceProtocl, bMaxPacketSize )
0xD8 , 0x04 ,      ( idVendor [low high] )
0x02 , 0x00 ,      ( idProduct [low high] )
0x00 , 0x00 ,      ( bcdDevice [low high] )
0x01 , 0x02 ,	   ( iManufacturer, iProduct )
0x00 , NUM_CONFIGURATIONS , ( iSerialNumber [none], bNumConfigurations )

( Configuration1 )
0x09 , CONFIGURATION ,  ( bLength, bDescriptorType )
0x22 , 0x00 ,	        ( wTotalLength LH )
NUM_INTERFACES , 0x01 ,	( bNumInterfaces, bConfigurationValue )
0x00 , 0xA0 ,		( iConfiguration [none], bmAttributes )
0x32 , 0x09 ,		( bMaxPower [100 mA]  bLength [Interface1 descriptor starts here] )
INTERFACE , 0x00 ,	( bDescriptorType, bInterfaceNumber )
0x00 , 0x01 , 		( bAlternateSetting, bNumEndpoints [excluding EP0] )
0x03 , 0x01 , 		( bInterfaceClass [HID code], bInterfaceSubClass [Boot subclass] )
0x01 , 0x00 ,		( bInterfaceProtocol [Keyboard protocol], iInterface [none] )

( HID1 )
0x09 , HID ,		( bLength, bDescriptorType )
0x00 , 0x01 , 		( bcdHID LH )
0x00 , 0x01 ,		( bCountryCode [none], bNumDescriptors )
REPORT , 0x3F ,		( bDescriptorType, wDescriptorLength [low byte] )
0x00 , 0x07 ,		( wDescriptorLength [high byte], bLength [Endpoint1 descritor starts here] )
ENDPOINT , 0x81 ,	( bDescriptorType, bEndpointAddress [EP1 IN] )
0x03 , 0x08 ,		( bmAttributes [Interrupt], wMaxPacketSize [low byte] )
0x00 , 0x0A ,		( wMaxPacketSize [high byte], bInterval [10 ms] )

( Report1 ) 
0x05 , 0x01 ,			( Usage Page [Generic Desktop], )
0x09 , 0x06 ,			( Usage [Keyboard], )
0xA1 , 0x01 ,			( Collection [Application], )
0x05 , 0x07 ,			(     Usage Page [Key Codes] )
0x19 , 0xE0 ,			(     Usage Minimum [224], )
0x29 , 0xE7 ,			(     Usage Maximum [231], )
0x15 , 0x00 ,			(     Logical Minimum [0], )
0x25 , 0x01 ,			(     Logical Maximum [1], )
0x75 , 0x01 ,			(     Report Size [1], )
0x95 , 0x08 ,			(     Report Count [8], )
0x81 , 0x02 ,   		(     Input [Data, Variable, Absolute],   ( Modifier byte )
0x95 , 0x01 ,			(     Report Count [1], )
0x75 , 0x08 ,			(     Report Size [8], )
0x81 , 0x01 ,			(     Input [Constant],                   ( Reserved byte )
0x95 , 0x05 ,			(     Report Count [5], )
0x75 , 0x01 ,			(     Report Size [1], )
0x05 , 0x08 ,			(     Usage Page [Page# for LEDs], )
0x19 , 0x01 ,			(     Usage Minimum [1], )
0x29 , 0x05 ,			(     Usage Maxmimum [5], )
0x91 , 0x02 ,			(     Output [Data, Variable, Absolute],  ( LED report )
0x95 , 0x01 ,			(     Report Count [1], )
0x75 , 0x03 ,			(     Report Size [3], )
0x91 , 0x01 ,			(     Output [Constant],                  ( LED report padding )
0x95 , 0x06 ,			(     Report Count [6], )
0x75 , 0x08 ,			(     Report Size [8], )
0x15 , 0x00 ,			(     Logical Minimum [0], )
0x25 , 0x65 ,			(     Logical Maximum [101], )
0x05 , 0x07 ,			(     Usage Page [Key Codes], )
0x19 , 0x00 ,			(     Usage Minimum [0], )
0x29 , 0x65 ,			(     Usage Maximum [101], )
0x81 , 0x00 ,			(     Input [Data, Array],                ( Key arrays [6 bytes] )
0xC0				( End Collection )

( string0 )
0x4 , STRING , ( bLength, bDescriptorType )
0x09 , 0x04 ,  ( wLANGID[0] lo hi )

( string1 )
0x??, STRING ,
"Microchip Technology, Inc."

( string2 )
0x??, STRING ,
S" ENGR 2210 PIC18F2455 USB Keyboard"

( interrupt flags )
macro
: trn?   UIR TRNFI   high? ;

    
: service-usb

    UIR UERRIF  high? if
	0 UEIR ! exit then      ( ignore error )

    UIR SOFIF   high? if
	UIR SOFIF low exit then ( soft detach ??? )

    UIR IDLEIF  high? if
	UIR IDLEIF low
	UCON SUSPND high
	exit then

    UIR ACTVIF  high? if
	UIR ACTVIF low
	UCON SUSPND low
	exit then

    UIR STALLIF high? if
	UIR STALLIF low exit then ( ignore stall )

    UIR URSTIF  high? if
	4 for UIR TRNIF low next ( clear TRNIF four times to clear out the USTAT FIFO )
	( stuff )
	exit then

    UIR TRNIF high? if
	( get buffer and handle:
	- setup token
	- in token
	- out token )
	exit then ;
