( -*- forth -*- )

( 12f synth controller for 18f )




target incremental

( port config )
: on   TRISA 0 low PORTA 0 high ;
: off  TRISA 0 low PORTA 0 low ;


( duty cycles )
: long  [ 10000000 2 * 3 / 8000 / 3 / 20 / ]L for 20 for next next ;
: short [ 10000000 1 * 3 / 8000 / 3 / 20 / ]L for 20 for next next ;

( the signal feeding the synth provides power, clock and data.
signal frequency is 8000Hz, with 3 bits per cycle:

110 = logic 1
100 = logic 0

this is equivalent to pwm, where the duty cycle 2/3 or 1/3 carries
the data bit. another way to see it is standard asynchronous
comm, 1 start, 1 data, 1 stop bit.

0->1 transition is clock edge
1->0 transition is data edge

these assume state is on, leave state on )
: cycle1 long  off short on ; 
: cycle0 short off long on ; 


( send a byte
on top of the pwm bit transfer there is a logical, asynchronous
byte channel. the format is

111...0dddddddd1....111

1    is idle
1->0 is start bit
d->1 is stop bit

)

: send
    cycle0 ( start bit )
    8 for
    WREG 0 high? if
	cycle1 else cycle0 then
    >> next drop
    cycle1 ( stop bit ) ;

( pilot tone
in theory 1 cycle should be enough when switching from
self clocked to external sync, but we send a little more because
1 don't work. doesn't seem to be a software bug.. )

( eeprom programming pause
specs say maximum erase/write cycle is 6 ms.
after reception of program byte, we pause before sending
more data. let's choose pilot and program delay the same. )

( 6.25 ms )
: wait 50 for cycle1 next ;
: pilot wait ;




( synth XTs )

( [ 10 | xx | dddd ] )
: note0! 0x24 send ;
: note1! 0x25 send ;
: note2! 0x26 send ;
: delay! 0x27 send ;

( [ 1110 | xxx | d )
: burst!  0x28 send ;
: octave! 0x29 send ;
: chirp!  0x2a send ;
: range!  0x2b send ;

( [ 11110 | xxx ] )
: period0! 0x34 send ;
: period1! 0x35 send ;
: period2! 0x36 send ;
: >engine  0x37 send ;
: engine!  0x38 send ;
: >trans   0x39 send ;

( [ 111110 | xx ] )
: rs>ip       0x3c send ;
: drop-engine 0x3d send ;
: swap-engine 0x3e send ;

( 0xFC [ 111111 | xx ] )
: pwm         0x40 send ;



( forth XTs )
: reset       0x00 send ;  ( reset chip )
: forth       0x01 send ;  ( enter tight forth loop )
: initsynth   0x02 send ;  ( initialize synth )
: sequencer   0x03 send ;  ( re-enter sequencer )
: read        0x4c send ;  ( load next byte on ds )
: (@)         0x4d send ;  ( fetch )
: (!)         0x4e send ;  ( store )
: bye         0x4f send ;  ( exit forth interpreter )
: ee@         0x50 send ;  ( fetch from eeprom )
: ee!         0x51 send ;  ( store to eeprom )
: program     0x52 send ;  ( program a sequence of bytes: n start -- )

( utility )
: >d          read send ;
: f-osc0!     not >d period0! ; ( these set the f registers directly )
: f-osc1!     not >d period1! ;
: f-osc2!     not >d period2! ;
: engine      >d >engine ;


: ports       0 TRISA ! 0 TRISB ! 0x80 TRISC ! ;
: debug       pilot 0x81 engine -1 f-osc0! -8 f-osc1!  ;
: debugdata   pilot 0x01 engine -1 f-osc0! -8 f-osc1!  ;
: synth!      >x pilot >d x> >d ee! wait ;
: restart     pilot initsynth sequencer  ;

: rst         pilot reset  ;
: frt         pilot forth  ;
: seq         pilot sequencer  ;

( upload
using buffered transfer:
* host forth transfers to controller
* controller transfers to target

using bottom half of bank 1

)

( transfer buffer to target )
: transfer    pilot forth 10 for wait next dup >d 0 >d program [ 0x100 ] seta for @a+ send wait next ;

: r1 0 for 0 for cycle1 next next ;
: r0 0 for 0 for cycle0 next next ;


( TODO: multisync
* interrupt based bit bang
* global sync transmit
* individual program upload
)


