( -*- forth -*- )

( example app for 14bit forth. tested on 12f629 )
( simple sound engine
runs at 8 khz.
this leaves us 125 cycles to do our magic.


3 level clocks:

synth:   8000khz [timer0: 125]
control: 40hz    /200
note:    8hz     /5 [120 bpm ]
bar:     .5hz    /16

)
load stc14/stc.f

target

( frequency division counters )
variable p-control ( / 200 )
variable p-note    ( / 5 )
variable p-bar     ( / 16 )
variable p-osc     ( variable )

( frequency counter: 4000 / period )
variable period

host
( frequency division factors )
200 neg constant CONTROL
5   neg constant NOTE
16  neg constant BAR

: hz  ( freq -- period ) 4000 swap / neg ;

macro

: control+ p-control count+ ;
: note+    p-note    count+ ;
: bar+     p-bar     count+ ;
: osc+     p-osc     count+ ;
: on       GPIO 5 high ;
: off      GPIO 5 low ;

target



: 8000hz
    INTCON T0IF low        ( clear interrupt bit )
    INTCON T0IE high       ( enable timer interrupt )
    [ 125 not ]L TMR0 ! ;  ( set timer value )


( oscillator )
sequence : toggle on | off ;
: osc     toggle period @ p-osc ! ;

( synth : variable frequency )
: synth   osc+ z? pass osc ; 

( interrupt handlers:
if a frequency division counter is 0, control
is passed to the lower level )

: bar
    [ BAR ]L p-bar ! iret ;

: note
    [ NOTE ]L p-note !
    bar+ z? pass bar iret ; 

: control
    -1   period +!
    0xfe period or!
    [ CONTROL ]L p-control !
    note+ z? pass note iret ;

interrupt : sample
    8000hz synth
    control+ z? pass control iret ;



: timer0
    bank1 clrwdt
    0x88 OPTION_REG ! ( pullup off, prescaler off )
    bank0 ;

application : main
    digital                    ( all digital io )
    bank1 0xf TRISIO ! bank0   ( 5 = out, rest = in )
    timer0
    interrupts
    8000hz
    [ 200 hz ]L period !
    begin again ;


there 2 / .x cr
savehex synth14.hex
