( -*- forth -*- )

( example app for 14bit forth. tested on 12f629 )
load stc14/stc.f

macro 
: on      GPIO 5 high ;
: off     GPIO 5 low ;

target

( busy delays )
: 1msec   250 begin 1- 0? until drop ;
: 10msec  10  begin 1msec 1- 0? until drop ;
: delay   begin 10msec 1- 0? until drop ;

: flash   on 5 delay off 25 delay ;
: pause   100 delay ;

: error   flash flash flash pause error ;

( flash sequence: -- units = 10msec 
  this is a task run from the timer interrupt handler )
sequence : task
    on 100 | off 100 ;
    

: 100hz
    INTCON T0IF low        ( clear interrupt bit )
    INTCON T0IE high       ( enable timer interrupt )
    [ 39 not ]L TMR0 ! ;   ( set timer value )

variable ticks
: timer-int
    100hz ticks @ 1- z?    ( decrement ticks )
    if drop task           ( if zero, get nb ticks from task )
    then ticks ! iret ;    ( update var & return )

( interrupt dispatcher )
interrupt : int
    bank0 INTCON T0IF high? pass timer-int
    error ;

: setup
    bank0 
    0 GPIO !
    7 CMCON !         ( digital io )
    bank1
    0 ANSEL !         ( disable analog inputs )
    0xf TRISIO !      ( 5 = out, rest = in )
    clrwdt
    0x87 OPTION_REG ! ( pullup off, prescaler TMR0 256x )
    bank0
    INTCON PEIE high  ( enable peripherial interrupts )
    INTCON GIE high   ( global interrupt enable )
;



application : main
    setup
    flash flash pause ( startup sequence )
    100hz 1 ticks !   ( init timer )
    begin again ;     ( mainloop )


there 2 / .x cr
savehex timer14.hex
