( -*- forth -*- )
( wisp628 interfacing )
host




: check [c] ? = if e-wisp then ;


: w out in ;                               ( send a byte to wisp + get ack )

( : w dup emit ; )

: <w>  c w ;
: [w]_ [c]_ w_ ;

: digits for dup 28 >> >hex w drop 4 << next  ;
: 1>w  24 << 2 digits drop ;
: 2>w  16 << 4 digits drop ;
: 3>w  8  << 6 digits drop ;
: >w   2>w ;

: 1w>  0 2 for 4 << [w] N hex> or next ;
: 2w>  0 4 for 4 << [w] N hex> or next ;
: w>   2w> ;

: jump  3>w [w] M check ;
: read  [w] R check w> ;
: 2write 2>w [w] W check ;
: 1write 1>w [w] W check ;
: inc   [w] I check ;
: +inc  for inc next ;
: write 2write ;


( change baud rate )
: p( >w [c] p out ;
: )c baud in dup emit check iflush ;

: 9k6   0x04 p(   9600 )c ;
: 19k2  0x14 p(  19200 )c ;
: 38k4  0x24 p(  38400 )c ;
: 57k6  0x34 p(  57600 )c ;
: 115k2 0x44 p( 115200 )c ;

    





( burn is defined in target specific code )

variable _algo
(
0x00 16x84, 16f62x and 16f87x
0x10 16f7x
0x20 16f87xA
0x30 18fxxx   [FIXME: writes 8 at a time]
0x40 12fxxx
)

: (program) _algo h@ or >w [w] X check ; ( send a program command, combine with algo select )
: algo 0xf and 4 << _algo h! ;           ( set read/write algo )

( wisp commands. currently specific for 12f629 )
: code     0xc (program) ;
: fuses    0xf (program) ;
: eeprom   0xd (program) ;
: erase    0xe (program) ;

: .id      8 for [w] n emit next ;
: deviceid fuses 6 +inc read ;

: wispid
    [w] t check .id
    [w] v check .id cr ;

: hello     0 >w [w] h ;
: attention
    iflush 19200 baud break
    in drop ( after break, we get a 0 ??? )
    hello ;
    

: wisp attention wispid ;

( generic wisp commands )
: go        attention 0 >w [w] g ;
: connect   attention 0x13 >w [c] p out 200000 usleep iflush ; ( we don't get an ack here )

( utility code )
: write-tick     dup t@ write inc 2 + ;
: progress       0xd emit dup 2 / .x space ;
: write-buffer   0 there 2 / for progress write-tick next drop ;
: verify-tick    dup t@ read = not if [c] ! emit cr then inc  2 + ;
: verify-buffer  0 there 2 / for progress verify-tick next drop ;
: verify         code verify-buffer ;
: ee-jump        dup if for inc next else drop then ;

: ee@            eeprom ee-jump read go ;
: ee!            eeprom ee-jump write go ; 

